@@ -3,7 +3,7 @@
    Dynamic DNS updates. */
 
 /*
- * Copyright (c) 2004-2007 by Internet Systems Consortium, Inc. ("ISC")
+ * Copyright (c) 2004 by Internet Systems Consortium, Inc. ("ISC")
  * Copyright (c) 2000-2003 by Internet Software Consortium
  *
  * Permission to use, copy, modify, and distribute this software for any
@@ -34,7 +34,7 @@
 
 #ifndef lint
 static char copyright[] =
-"$Id: ddns.c,v 1.23.10.3 2007/04/27 22:48:10 each Exp $ Copyright (c) 2004-2007 Internet Systems Consortium.  All rights reserved.\n";
+"$Id: ddns.c,v 1.15.2.14 2004/06/17 20:54:40 dhankins Exp $ Copyright (c) 2004 Internet Systems Consortium.  All rights reserved.\n";
 #endif /* not lint */
 
 #include "dhcpd.h"
@@ -231,9 +231,8 @@ int ddns_updates (struct packet *packet,
 	int result = 0;
 	isc_result_t rcode1 = ISC_R_SUCCESS, rcode2 = ISC_R_SUCCESS;
 	int server_updates_a = 1;
-	int server_updates_ptr = 1;
 	struct buffer *bp = (struct buffer *)0;
-	int ignorep = 0, client_ignorep = 0;
+	int ignorep = 0;
 
 	if (ddns_update_style != 2)
 		return 0;
@@ -242,7 +241,6 @@ int ddns_updates (struct packet *packet,
 	if (lease -> ip_addr . len != 4)
 		return 0;
 
-	memset(&d1, 0, sizeof(d1));
 	memset (&ddns_hostname, 0, sizeof (ddns_hostname));
 	memset (&ddns_domainname, 0, sizeof (ddns_domainname));
 	memset (&old_ddns_fwd_name, 0, sizeof (ddns_fwd_name));
@@ -252,11 +250,13 @@ int ddns_updates (struct packet *packet,
 
 	/* If we are allowed to accept the client's update of its own A
 	   record, see if the client wants to update its own A record. */
-	if (!(oc = lookup_option(&server_universe, state->options,
-				 SV_CLIENT_UPDATES)) ||
-	    evaluate_boolean_option_cache(&client_ignorep, packet, lease, NULL,
-					  packet->options, state->options,
-					  &lease->scope, oc, MDL)) {
+	if (!(oc = lookup_option (&server_universe, state -> options,
+				  SV_CLIENT_UPDATES)) ||
+	    evaluate_boolean_option_cache (&ignorep, packet, lease,
+					   (struct client_state *)0,
+					   packet -> options,
+					   state -> options,
+					   &lease -> scope, oc, MDL)) {
 		/* If there's no fqdn.no-client-update or if it's
 		   nonzero, don't try to use the client-supplied
 		   XXX */
@@ -315,6 +315,8 @@ int ddns_updates (struct packet *packet,
 	/*
 	 * Compute the name for the A record.
 	 */
+	s1 = s2 = 0;
+
 	oc = lookup_option (&server_universe, state -> options,
 			    SV_DDNS_HOST_NAME);
 	if (oc)
@@ -323,8 +325,6 @@ int ddns_updates (struct packet *packet,
 					    packet -> options,
 					    state -> options,
 					    &lease -> scope, oc, MDL);
-	else
-		s1 = 0;
 
 	oc = lookup_option (&server_universe, state -> options,
 			    SV_DDNS_DOMAIN_NAME);
@@ -334,8 +334,6 @@ int ddns_updates (struct packet *packet,
 					    packet -> options,
 					    state -> options,
 					    &lease -> scope, oc, MDL);
-	else
-		s2 = 0;
 
 	if (s1 && s2) {
 		if (ddns_hostname.len + ddns_domainname.len > 253) {
@@ -400,8 +398,7 @@ int ddns_updates (struct packet *packet,
 				/* Now, install the DDNS data the new way. */
 				goto in;
 			}
-		} else
-			data_string_forget(&ddns_dhcid, MDL);
+		}
 
 		/* See if the administrator wants to do updates even
 		   in cases where the update already appears to have been
@@ -416,12 +413,15 @@ int ddns_updates (struct packet *packet,
 			result = 1;
 			goto noerror;
 		}
-	/* If there's no "ddns-fwd-name" on the lease record, see if
-	 * there's a ddns-client-fqdn indicating a previous client
-	 * update (if it changes, we need to adjust the PTR).
-	 */
-	} else if (find_bound_string(&old_ddns_fwd_name, lease->scope,
-				     "ddns-client-fqdn")) {
+	}
+
+	/* If there's no ddns-fwd-name on the lease, see if there's
+	   a ddns-client-fqdn, indicating a prior client FQDN update.
+	   If there is, and if we're still doing the client update,
+	   see if the name has changed.   If it hasn't, don't do the
+	   PTR update. */
+	if (find_bound_string (&old_ddns_fwd_name,
+			       lease -> scope, "ddns-client-fqdn")) {
 		/* If the name is not different, no need to update
 		   the PTR record. */
 		if (old_ddns_fwd_name.len == ddns_fwd_name.len &&
@@ -455,6 +455,7 @@ int ddns_updates (struct packet *packet,
 	 * Compute the RR TTL.
 	 */
 	ddns_ttl = DEFAULT_DDNS_TTL;
+	memset (&d1, 0, sizeof d1);
 	if ((oc = lookup_option (&server_universe, state -> options,
 				 SV_DDNS_TTL))) {
 		if (evaluate_option_cache (&d1, packet, lease,
@@ -468,17 +469,6 @@ int ddns_updates (struct packet *packet,
 		}
 	}
 
-	/* CC: see if we are configured NOT to do reverse ptr updates
-        */
-	if ((oc = lookup_option (&server_universe, state -> options,
-				 SV_DO_REVERSE_UPDATES)) &&
-	    !evaluate_boolean_option_cache (&ignorep, packet, lease,
-					    (struct client_state *)0,
-					    packet -> options,
-					    state -> options,
-					    &lease -> scope, oc, MDL)) {
-		server_updates_ptr = 0;
-	}
 
 	/*
 	 * Compute the reverse IP name.
@@ -491,10 +481,8 @@ int ddns_updates (struct packet *packet,
 					    packet -> options,
 					    state -> options,
 					    &lease -> scope, oc, MDL);
-	else
-		s1 = 0;
 
-	if (s1 && (d1.len > 238)) {
+	if (d1.len > 238) {
 		log_error ("ddns_update: Calculated rev domain name too long.");
 		s1 = 0;
 		data_string_forget (&d1, MDL);
@@ -530,6 +518,7 @@ int ddns_updates (struct packet *packet,
 	 * If we are updating the A record, compute the DHCID value.
 	 */
 	if (server_updates_a) {
+		memset (&ddns_dhcid, 0, sizeof ddns_dhcid);
 		if (lease -> uid && lease -> uid_len)
 			result = get_dhcid (&ddns_dhcid,
 					    DHO_DHCP_CLIENT_IDENTIFIER,
@@ -555,25 +544,11 @@ int ddns_updates (struct packet *packet,
 	/*
 	 * Perform updates.
 	 */
-	if (ddns_fwd_name.len && ddns_dhcid.len) {
-		unsigned conflict;
-
-		oc = lookup_option(&server_universe, state->options,
-				   SV_DDNS_CONFLICT_DETECT);
-		if (!oc ||
-		    evaluate_boolean_option_cache(&ignorep, packet, lease,
-						  NULL, packet->options,
-						  state->options,
-						  &lease->scope, oc, MDL))
-			conflict = 1;
-		else
-			conflict = 0;
-
+	if (ddns_fwd_name.len && ddns_dhcid.len)
 		rcode1 = ddns_update_a (&ddns_fwd_name, lease -> ip_addr,
-					&ddns_dhcid, ddns_ttl, 0, conflict);
-	}
-
-	if (rcode1 == ISC_R_SUCCESS && server_updates_ptr) {
+					&ddns_dhcid, ddns_ttl, 0);
+	
+	if (rcode1 == ISC_R_SUCCESS) {
 		if (ddns_fwd_name.len && ddns_rev_name.len)
 			rcode2 = ddns_update_ptr (&ddns_fwd_name,
 						  &ddns_rev_name, ddns_ttl);
@@ -591,151 +566,59 @@ int ddns_updates (struct packet *packet,
 				       &ddns_dhcid);
 	}
 
-	if (rcode2 == ISC_R_SUCCESS && server_updates_ptr) {
+	if (rcode2 == ISC_R_SUCCESS) {
 		bind_ds_value (&lease -> scope, "ddns-rev-name",
 			       &ddns_rev_name);
 	}
 
-      noerror:
-	/*
-	 * If fqdn-reply option is disabled in dhcpd.conf, then don't
-	 * send the client an FQDN option at all, even if one was requested.
-	 * (WinXP clients allegedly misbehave if the option is present,
-	 * refusing to handle PTR updates themselves).
-	 */
-	if ((oc = lookup_option (&server_universe, state->options,
-  				 SV_FQDN_REPLY)) &&
-  	    !evaluate_boolean_option_cache (&ignorep, packet, lease,
-  					    (struct client_state *)0,
-  					    packet->options,
-  					    state->options,
-  					    &lease->scope, oc, MDL)) {
-  	    	goto badfqdn;
-
-	/* If we're ignoring client updates, then we tell a sort of 'white
-	 * lie'.  We've already updated the name the server wants (per the
-	 * config written by the server admin).  Now let the client do as
-	 * it pleases with the name they supplied (if any).
-	 *
-	 * We only form an FQDN option this way if the client supplied an
-	 * FQDN option that had FQDN_SERVER_UPDATE set false.
-	 */
-	} else if (client_ignorep &&
-	    (oc = lookup_option(&fqdn_universe, packet->options,
-				FQDN_SERVER_UPDATE)) &&
-	    !evaluate_boolean_option_cache(&ignorep, packet, lease, NULL,
-					   packet->options, state->options,
-					   &lease->scope, oc, MDL)) {
-		oc = lookup_option(&fqdn_universe, packet->options, FQDN_FQDN);
-		if (oc && evaluate_option_cache(&d1, packet, lease, NULL,
-						packet->options, state->options,
-						&global_scope, oc, MDL)) {
-			if (d1.len == 0 ||
-			    !buffer_allocate(&bp, d1.len + 5, MDL))
-				goto badfqdn;
-
-			/* Server pretends it is not updating. */
-			bp->data[0] = 0;
-			if (!save_option_buffer(&fqdn_universe, state->options,
-						bp, &bp->data[0], 1,
-						FQDN_SERVER_UPDATE, 0))
-				goto badfqdn;
-
-			/* Client is encouraged to update. */
-			bp->data[1] = 0;
-			if (!save_option_buffer(&fqdn_universe, state->options,
-						bp, &bp->data[1], 1,
-						FQDN_NO_CLIENT_UPDATE, 0))
-				goto badfqdn;
-
-			/* Use the encoding of client's FQDN option. */
-			oc = lookup_option(&fqdn_universe, packet->options,
-					   FQDN_ENCODED);
-			if (oc && evaluate_boolean_option_cache(&ignorep,
-							packet, lease, NULL,
-							packet->options,
-							state->options,
-							&lease->scope, oc,
-							MDL))
-				bp->data[2] = 1; /* FQDN is encoded. */
-			else
-				bp->data[2] = 0; /* FQDN is not encoded. */
-
-			if (!save_option_buffer(&fqdn_universe, state->options,
-						bp, &bp->data[2], 1,
-						FQDN_ENCODED, 0))
-				goto badfqdn;
-
-			/* Current FQDN drafts indicate 255 is mandatory. */
-			bp->data[3] = 255;
-			if (!save_option_buffer(&fqdn_universe, state->options,
-						bp, &bp->data[3], 1,
-						FQDN_RCODE1, 0))
-				goto badfqdn;
-
-			bp->data[4] = 255;
-			if (!save_option_buffer(&fqdn_universe, state->options,
-						bp, &bp->data[4], 1,
-						FQDN_RCODE2, 0))
-				goto badfqdn;
-
-			/* Copy in the FQDN supplied by the client.  Note well
-			 * that the format of this option in the cache is going
-			 * to be in text format.  If the fqdn supplied by the
-			 * client is encoded, it is decoded into the option
-			 * cache when parsed out of the packet.  It will be
-			 * re-encoded when the option is assembled to be
-			 * transmitted if the client elects that encoding.
-			 */
-			memcpy(&bp->data[5], d1.data, d1.len);
-			if (!save_option_buffer(&fqdn_universe, state->options,
-						bp, &bp->data[5], 1,
-						FQDN_FQDN, 0))
-				goto badfqdn;
-
-			data_string_forget(&d1, MDL);
-		}
 	/* Set up the outgoing FQDN option if there was an incoming
-	 * FQDN option.  If there's a valid FQDN option, there MUST
-	 * be an FQDN_SERVER_UPDATES suboption, it's part of the fixed
-	 * length head of the option contents, so we test the latter
-	 * to detect the presence of the former.
-	 */
-	} else if ((oc = lookup_option(&fqdn_universe, packet->options,
-				       FQDN_ENCODED)) &&
-		   buffer_allocate(&bp, ddns_fwd_name.len + 5, MDL)) {
+	   FQDN option.  If there's a valid FQDN option, there should
+	   be an FQDN_ENCODED suboption, so we test the latter to
+	   detect the presence of the former. */
+      noerror:
+	if ((oc = lookup_option (&fqdn_universe,
+				 packet -> options, FQDN_ENCODED))
+	    && buffer_allocate (&bp, ddns_fwd_name.len + 5, MDL)) {
 		bp -> data [0] = server_updates_a;
 		if (!save_option_buffer (&fqdn_universe, state -> options,
 					 bp, &bp -> data [0], 1,
-					 FQDN_SERVER_UPDATE, 0))
+					 &fqdn_options [FQDN_SERVER_UPDATE],
+					 0))
 			goto badfqdn;
 		bp -> data [1] = server_updates_a;
 		if (!save_option_buffer (&fqdn_universe, state -> options,
 					 bp, &bp -> data [1], 1,
-					 FQDN_NO_CLIENT_UPDATE, 0))
+					 &fqdn_options [FQDN_NO_CLIENT_UPDATE],
+					 0))
 			goto badfqdn;
-
 		/* Do the same encoding the client did. */
-		if (evaluate_boolean_option_cache(&ignorep, packet, lease,
-						  NULL, packet->options,
-						  state->options,
-						  &lease->scope, oc, MDL))
+		oc = lookup_option (&fqdn_universe, packet -> options,
+				    FQDN_ENCODED);
+		if (oc &&
+		    evaluate_boolean_option_cache (&ignorep, packet, lease,
+						   (struct client_state *)0,
+						   packet -> options,
+						   state -> options,
+						   &lease -> scope, oc, MDL))
 			bp -> data [2] = 1;
 		else
 			bp -> data [2] = 0;
 		if (!save_option_buffer (&fqdn_universe, state -> options,
 					 bp, &bp -> data [2], 1,
-					 FQDN_ENCODED, 0))
+					 &fqdn_options [FQDN_ENCODED],
+					 0))
 			goto badfqdn;
 		bp -> data [3] = isc_rcode_to_ns (rcode1);
 		if (!save_option_buffer (&fqdn_universe, state -> options,
 					 bp, &bp -> data [3], 1,
-					 FQDN_RCODE1, 0))
+					 &fqdn_options [FQDN_RCODE1],
+					 0))
 			goto badfqdn;
 		bp -> data [4] = isc_rcode_to_ns (rcode2);
 		if (!save_option_buffer (&fqdn_universe, state -> options,
 					 bp, &bp -> data [4], 1,
-					 FQDN_RCODE2, 0))
+					 &fqdn_options [FQDN_RCODE2],
+					 0))
 			goto badfqdn;
 		if (ddns_fwd_name.len) {
 		    memcpy (&bp -> data [5],
@@ -743,7 +626,8 @@ int ddns_updates (struct packet *packet,
 		    if (!save_option_buffer (&fqdn_universe, state -> options,
 					     bp, &bp -> data [5],
 					     ddns_fwd_name.len,
-					     FQDN_FQDN, 0))
+					     &fqdn_options [FQDN_FQDN],
+					     0))
 			goto badfqdn;
 		}
 	}
@@ -753,15 +637,14 @@ int ddns_updates (struct packet *packet,
 	/*
 	 * Final cleanup.
 	 */
-	data_string_forget(&d1, MDL);
-	data_string_forget(&ddns_hostname, MDL);
-	data_string_forget(&ddns_domainname, MDL);
-	data_string_forget(&old_ddns_fwd_name, MDL);
-	data_string_forget(&ddns_fwd_name, MDL);
-	data_string_forget(&ddns_rev_name, MDL);
-	data_string_forget(&ddns_dhcid, MDL);
+	data_string_forget (&ddns_hostname, MDL);
+	data_string_forget (&ddns_domainname, MDL);
+	data_string_forget (&old_ddns_fwd_name, MDL);
+	data_string_forget (&ddns_fwd_name, MDL);
+	data_string_forget (&ddns_rev_name, MDL);
+	data_string_forget (&ddns_dhcid, MDL);
 	if (bp)
-		buffer_dereference(&bp, MDL);
+		buffer_dereference (&bp, MDL);
 
 	return result;
 }
@@ -796,8 +679,6 @@ int ddns_removals (struct lease *lease)
 	if (!resolver_inited) {
 		minires_ninit (&resolver_state);
 		resolver_inited = 1;
-		resolver_state.retrans = 1;
-		resolver_state.retry = 1;
 	}
 
 	/* We need the fwd name whether we are deleting both records or just
