@@ -3,7 +3,7 @@
    Server-specific in-memory database support. */
 
 /*
- * Copyright (c) 2004-2008 by Internet Systems Consortium, Inc. ("ISC")
+ * Copyright (c) 2004 by Internet Systems Consortium, Inc. ("ISC")
  * Copyright (c) 1996-2003 by Internet Software Consortium
  *
  * Permission to use, copy, modify, and distribute this software for any
@@ -34,7 +34,7 @@
 
 #ifndef lint
 static char copyright[] =
-"$Id: mdb.c,v 1.83.16.10 2008/01/22 19:02:51 dhankins Exp $ Copyright (c) 2004-2008 Internet Systems Consortium.  All rights reserved.\n";
+"$Id: mdb.c,v 1.67.2.19 2004/06/10 17:59:56 dhankins Exp $ Copyright (c) 2004 Internet Systems Consortium.  All rights reserved.\n";
 #endif /* not lint */
 
 #include "dhcpd.h"
@@ -45,65 +45,12 @@ struct shared_network *shared_networks;
 host_hash_t *host_hw_addr_hash;
 host_hash_t *host_uid_hash;
 host_hash_t *host_name_hash;
-lease_id_hash_t *lease_uid_hash;
-lease_ip_hash_t *lease_ip_addr_hash;
-lease_id_hash_t *lease_hw_addr_hash;
-
-int numclasseswritten;
+lease_hash_t *lease_uid_hash;
+lease_hash_t *lease_ip_addr_hash;
+lease_hash_t *lease_hw_addr_hash;
 
 omapi_object_type_t *dhcp_type_host;
 
-isc_result_t enter_class(cd, dynamicp, commit)
-	struct class *cd;
-	int dynamicp;
-	int commit;
-{
-	if (!collections -> classes) {
-		/* A subclass with no parent is invalid. */
-		if (cd->name == NULL)
-			return ISC_R_INVALIDARG;
-
-		class_reference (&collections -> classes, cd, MDL);
-	} else if (cd->name != NULL) {	/* regular class */
-		struct class *c = 0;
-
-		if (find_class(&c, cd->name, MDL) != ISC_R_NOTFOUND) {
-			class_dereference(&c, MDL);
-			return ISC_R_EXISTS;
-		}
-		
-		/* Find the tail. */
-		for (c = collections -> classes;
-		     c -> nic; c = c -> nic)
-			/* nothing */ ;
-		class_reference (&c -> nic, cd, MDL);
-	}
-
-	if (dynamicp && commit) {
-		const char *name = cd->name;
-
-		if (name == NULL) {
-			name = cd->superclass->name;
-		}
-
-		write_named_billing_class ((const unsigned char *)name, 0, cd);
-		if (!commit_leases ())
-			return ISC_R_IOERROR;
-	}
-
-	return ISC_R_SUCCESS;
-}
-
-
-/* Variable to check if we're starting the server.  The server will init as
- * starting - but just to be safe start out as false to avoid triggering new
- * special-case code
- * XXX: There is actually a server_startup state...which is never entered...
- */
-#define SS_NOSYNC	1
-#define SS_QFOLLOW	2
-static int server_starting = 0;
-
 static int find_uid_statement (struct executable_statement *esp,
 			       void *vp, int condp)
 {
@@ -139,7 +86,7 @@ isc_result_t enter_host (hd, dynamicp, commit)
 	struct executable_statement *esp;
 
 	if (!host_name_hash) {
-		if (!host_new_hash(&host_name_hash, HOST_HASH_SIZE, MDL))
+		if (!host_new_hash (&host_name_hash, 0, MDL))
 			log_fatal ("Can't allocate host name hash");
 		host_hash_add (host_name_hash,
 			       (unsigned char *)hd -> name,
@@ -194,8 +141,7 @@ isc_result_t enter_host (hd, dynamicp, commit)
 
 	if (hd -> interface.hlen) {
 		if (!host_hw_addr_hash) {
-			if (!host_new_hash(&host_hw_addr_hash,
-					   HOST_HASH_SIZE, MDL))
+			if (!host_new_hash (&host_hw_addr_hash, 0, MDL))
 				log_fatal ("Can't allocate host/hw hash");
 		} else {
 			/* If there isn't already a host decl matching this
@@ -239,8 +185,7 @@ isc_result_t enter_host (hd, dynamicp, commit)
 		/* If there's no uid hash, make one; otherwise, see if
 		   there's already an entry in the hash for this host. */
 		if (!host_uid_hash) {
-			if (!host_new_hash(&host_uid_hash,
-					   HOST_HASH_SIZE, MDL))
+			if (!host_new_hash (&host_uid_hash, 0, MDL))
 				log_fatal ("Can't allocate host/uid hash");
 
 			host_hash_add (host_uid_hash,
@@ -286,30 +231,6 @@ isc_result_t enter_host (hd, dynamicp, commit)
 	return ISC_R_SUCCESS;
 }
 
-
-isc_result_t delete_class (cp, commit)
-	struct class *cp;
-	int commit;
-{
-	cp->flags |= CLASS_DECL_DELETED;
-
-	/* do the write first as we won't be leaving it in any data
-	   structures, unlike the host objects */
-	
-	if (commit) {
-		write_named_billing_class ((unsigned char *)cp->name, 0, cp);
-		if (!commit_leases ())
-			return ISC_R_IOERROR;
-	}
-	
-	unlink_class(&cp);		/* remove from collections */
-
-	class_dereference(&cp, MDL);
-
-	return ISC_R_SUCCESS;
-}
-
-
 isc_result_t delete_host (hd, commit)
 	struct host_decl *hd;
 	int commit;
@@ -533,6 +454,7 @@ void new_address_range (cfile, low, high, subnet, pool, lpchain)
 	struct lease **lpchain;
 {
 	struct lease *address_range, *lp, *plp;
+	struct iaddr net;
 	unsigned min, max, i;
 	char lowbuf [16], highbuf [16], netbuf [16];
 	struct shared_network *share = subnet -> shared_network;
@@ -548,33 +470,35 @@ void new_address_range (cfile, low, high, subnet, pool, lpchain)
 
 	/* Initialize the hash table if it hasn't been done yet. */
 	if (!lease_uid_hash) {
-		if (!lease_id_new_hash(&lease_uid_hash, LEASE_HASH_SIZE, MDL))
+		if (!lease_new_hash (&lease_uid_hash, 0, MDL))
 			log_fatal ("Can't allocate lease/uid hash");
 	}
 	if (!lease_ip_addr_hash) {
-		if (!lease_ip_new_hash(&lease_ip_addr_hash, LEASE_HASH_SIZE,
-				       MDL))
+		if (!lease_new_hash (&lease_ip_addr_hash, 0, MDL))
 			log_fatal ("Can't allocate lease/ip hash");
 	}
 	if (!lease_hw_addr_hash) {
-		if (!lease_id_new_hash(&lease_hw_addr_hash, LEASE_HASH_SIZE,
-				       MDL))
+		if (!lease_new_hash (&lease_hw_addr_hash, 0, MDL))
 			log_fatal ("Can't allocate lease/hw hash");
 	}
 
-	/* Make sure that high and low addresses are in this subnet. */
-	if (!addr_eq(subnet->net, subnet_number(low, subnet->netmask))) {
-		strcpy(lowbuf, piaddr(low));
-		strcpy(netbuf, piaddr(subnet->net));
-		log_fatal("bad range, address %s not in subnet %s netmask %s",
-			  lowbuf, netbuf, piaddr(subnet->netmask));
+	/* Make sure that high and low addresses are in same subnet. */
+	net = subnet_number (low, subnet -> netmask);
+	if (!addr_eq (net, subnet_number (high, subnet -> netmask))) {
+		strcpy (lowbuf, piaddr (low));
+		strcpy (highbuf, piaddr (high));
+		strcpy (netbuf, piaddr (subnet -> netmask));
+		log_fatal ("Address range %s to %s, netmask %s spans %s!",
+		       lowbuf, highbuf, netbuf, "multiple subnets");
 	}
 
-	if (!addr_eq(subnet->net, subnet_number(high, subnet->netmask))) {
-		strcpy(highbuf, piaddr(high));
-		strcpy(netbuf, piaddr(subnet->net));
-		log_fatal("bad range, address %s not in subnet %s netmask %s",
-			  highbuf, netbuf, piaddr(subnet->netmask));
+	/* Make sure that the addresses are on the correct subnet. */
+	if (!addr_eq (net, subnet -> net)) {
+		strcpy (lowbuf, piaddr (low));
+		strcpy (highbuf, piaddr (high));
+		strcpy (netbuf, piaddr (subnet -> netmask));
+		log_fatal ("Address range %s to %s not on net %s/%s!",
+		       lowbuf, highbuf, piaddr (subnet -> net), netbuf);
 	}
 
 	/* Get the high and low host addresses... */
@@ -617,7 +541,7 @@ void new_address_range (cfile, low, high, subnet, pool, lpchain)
 #endif
 		lp -> ip_addr = ip_addr (subnet -> net,
 					 subnet -> netmask, i + min);
-		lp -> starts = MIN_TIME;
+		lp -> starts = lp -> timestamp = MIN_TIME;
 		lp -> ends = MIN_TIME;
 		subnet_reference (&lp -> subnet, subnet, MDL);
 		pool_reference (&lp -> pool, pool, MDL);
@@ -635,9 +559,9 @@ void new_address_range (cfile, low, high, subnet, pool, lpchain)
 				pool_reference (&lt -> pool, pool, MDL);
 			lease_dereference (&lt, MDL);
 		} else
-			lease_ip_hash_add(lease_ip_addr_hash,
-					  lp->ip_addr.iabuf, lp->ip_addr.len,
-					  lp, MDL);
+			lease_hash_add (lease_ip_addr_hash,
+					lp -> ip_addr.iabuf,
+					lp -> ip_addr.len, lp, MDL);
 		/* Put the lease on the chain for the caller. */
 		if (lpchain) {
 			if (*lpchain) {
@@ -833,9 +757,9 @@ void enter_lease (lease)
 		if (comp -> subnet)
 			subnet_reference (&lease -> subnet,
 					  comp -> subnet, MDL);
-		lease_ip_hash_delete(lease_ip_addr_hash,
-				     lease->ip_addr.iabuf, lease->ip_addr.len,
-				     MDL);
+		lease_hash_delete (lease_ip_addr_hash,
+				   lease -> ip_addr.iabuf,
+				   lease -> ip_addr.len, MDL);
 		lease_dereference (&comp, MDL);
 	}
 
@@ -852,8 +776,9 @@ void enter_lease (lease)
 		log_error ("lease %s: no subnet.", piaddr (lease -> ip_addr));
 		return;
 	}
-	lease_ip_hash_add(lease_ip_addr_hash, lease->ip_addr.iabuf,
-			  lease->ip_addr.len, lease, MDL);
+	lease_hash_add (lease_ip_addr_hash,
+			lease -> ip_addr.iabuf,
+			lease -> ip_addr.len, lease, MDL);
 }
 
 /* Replace the data in an existing lease with the data in a new lease;
@@ -867,11 +792,12 @@ int supersede_lease (comp, lease, commit, propogate, pimmediate)
 	int propogate;
 	int pimmediate;
 {
+	int enter_uid = 0;
+	int enter_hwaddr = 0;
 	struct lease *lp, **lq, *prev;
 	TIME lp_next_state;
-#if defined (FAILOVER_PROTOCOL)
-	int do_pool_check = 0;
 
+#if defined (FAILOVER_PROTOCOL)
 	/* We must commit leases before sending updates regarding them
 	   to failover peers.  It is, therefore, an error to set pimmediate
 	   and not commit. */
@@ -914,22 +840,29 @@ int supersede_lease (comp, lease, commit, propogate, pimmediate)
 
 	/* If there's a Unique ID, dissociate it from the hash
 	   table and free it if necessary. */
-	if (comp->uid) {
-		uid_hash_delete(comp);
-		if (comp->uid != comp->uid_buf) {
-			dfree(comp->uid, MDL);
-			comp->uid_max = 0;
-			comp->uid_len = 0;
+	if (comp -> uid) {
+		uid_hash_delete (comp);
+		enter_uid = 1;
+		if (comp -> uid != &comp -> uid_buf [0]) {
+			dfree (comp -> uid, MDL);
+			comp -> uid_max = 0;
+			comp -> uid_len = 0;
 		}
 		comp -> uid = (unsigned char *)0;
-	}
-
-	/* If there's a hardware address, remove the lease from its
-	 * old position in the hash bucket's ordered list.
-	 */
-	if (comp->hardware_addr.hlen)
-		hw_hash_delete(comp);
-
+	} else
+		enter_uid = 1;
+	
+	if (comp -> hardware_addr.hlen &&
+	    ((comp -> hardware_addr.hlen !=
+	      lease -> hardware_addr.hlen) ||
+	     memcmp (comp -> hardware_addr.hbuf,
+		     lease -> hardware_addr.hbuf,
+		     comp -> hardware_addr.hlen))) {
+		hw_hash_delete (comp);
+		enter_hwaddr = 1;
+	} else if (!comp -> hardware_addr.hlen)
+		enter_hwaddr = 1;
+	
 	/* If the lease has been billed to a class, remove the billing. */
 	if (comp -> billing_class != lease -> billing_class) {
 		if (comp -> billing_class)
@@ -1016,34 +949,26 @@ int supersede_lease (comp, lease, commit, propogate, pimmediate)
 		executable_statement_reference (&comp -> on_release,
 						lease -> on_release, MDL);
 	}
-
+	
 	/* Record the lease in the uid hash if necessary. */
-	if (comp->uid)
-		uid_hash_add(comp);
-
+	if (enter_uid && comp -> uid) {
+		uid_hash_add (comp);
+	}
+	
 	/* Record it in the hardware address hash if necessary. */
-	if (comp->hardware_addr.hlen)
-		hw_hash_add(comp);
-
-	comp->cltt = lease->cltt;
+	if (enter_hwaddr && lease -> hardware_addr.hlen) {
+		hw_hash_add (comp);
+	}
+	
 #if defined (FAILOVER_PROTOCOL)
-	comp->tstp = lease->tstp;
-	comp->tsfp = lease->tsfp;
-	comp->atsfp = lease->atsfp;
+	comp -> cltt = lease -> cltt;
+	comp -> tstp = lease -> tstp;
+	comp -> tsfp = lease -> tsfp;
 #endif /* FAILOVER_PROTOCOL */
-	comp->ends = lease->ends;
-	comp->next_binding_state = lease->next_binding_state;
+	comp -> ends = lease -> ends;
+	comp -> next_binding_state = lease -> next_binding_state;
 
       just_move_it:
-#if defined (FAILOVER_PROTOCOL)
-	/* Atsfp should be cleared upon any state change that implies
-	 * propogation wether supersede_lease was given a copy lease
-	 * structure or not (often from the pool_timer()).
-	 */
-	if (propogate)
-		comp->atsfp = 0;
-#endif /* FAILOVER_PROTOCOL */
-
 	if (!comp -> pool) {
 		log_error ("Supersede_lease: lease %s with no pool.",
 			   piaddr (comp -> ip_addr));
@@ -1054,12 +979,7 @@ int supersede_lease (comp, lease, commit, propogate, pimmediate)
 	switch (comp -> binding_state) {
 	      case FTS_FREE:
 		lq = &comp -> pool -> free;
-		if (!(comp->flags & RESERVED_LEASE))
-			comp->pool->free_leases--;
-
-#if defined(FAILOVER_PROTOCOL)
-		do_pool_check = 1;
-#endif
+		comp -> pool -> free_leases--;
 		break;
 
 	      case FTS_ACTIVE:
@@ -1078,12 +998,7 @@ int supersede_lease (comp, lease, commit, propogate, pimmediate)
 
 	      case FTS_BACKUP:
 		lq = &comp -> pool -> backup;
-		if (!(comp->flags & RESERVED_LEASE))
-			comp->pool->backup_leases--;
-
-#if defined(FAILOVER_PROTOCOL)
-		do_pool_check = 1;
-#endif
+		comp -> pool -> backup_leases--;
 		break;
 
 	      default:
@@ -1097,7 +1012,6 @@ int supersede_lease (comp, lease, commit, propogate, pimmediate)
 
 	/* Remove the lease from its current place in its current
 	   timer sequence. */
-	/* XXX this is horrid. */
 	prev = (struct lease *)0;
 	for (lp = *lq; lp; lp = lp -> next) {
 		if (lp == comp)
@@ -1106,13 +1020,14 @@ int supersede_lease (comp, lease, commit, propogate, pimmediate)
 	}
 
 	if (!lp) {
-		log_fatal("Lease with binding state %s not on its queue.",
-			  (comp->binding_state < 1 ||
-			   comp->binding_state > FTS_LAST)
-			  ? "unknown"
-			  : binding_state_names[comp->binding_state - 1]);
+		log_error ("Lease with binding state %s not on its queue.",
+			   (comp -> binding_state < 1 ||
+			    comp -> binding_state > FTS_LAST)
+			   ? "unknown"
+			   : binding_state_names [comp -> binding_state - 1]);
+		return 0;
 	}
-
+	
 	if (prev) {
 		lease_dereference (&prev -> next, MDL);
 		if (comp -> next) {
@@ -1162,10 +1077,8 @@ int supersede_lease (comp, lease, commit, propogate, pimmediate)
 	if (commit) {
 		if (!write_lease (comp))
 			return 0;
-		if ((server_starting & SS_NOSYNC) == 0) {
-			if (!commit_leases ())
-				return 0;
-		}
+		if (!commit_leases ())
+			return 0;
 	}
 
 #if defined (FAILOVER_PROTOCOL)
@@ -1174,8 +1087,6 @@ int supersede_lease (comp, lease, commit, propogate, pimmediate)
 		if (!dhcp_failover_queue_update (comp, pimmediate))
 			return 0;
 	}
-	if (do_pool_check && comp->pool->failover_peer)
-		dhcp_failover_pool_check(comp->pool);
 #endif
 
 	/* If the current binding state has already expired, do an
@@ -1208,9 +1119,7 @@ void make_binding_state_transition (struct lease *lease)
 	    ((
 #if defined (FAILOVER_PROTOCOL)
 		    peer &&
-		    (lease -> binding_state == FTS_EXPIRED ||
-		     (peer -> i_am == secondary &&
-		      lease -> binding_state == FTS_ACTIVE)) &&
+		    lease -> binding_state == FTS_EXPIRED &&
 		    (lease -> next_binding_state == FTS_FREE ||
 		     lease -> next_binding_state == FTS_BACKUP)) ||
 	     (!peer &&
@@ -1369,6 +1278,7 @@ int lease_copy (struct lease **lp,
 	lt -> ip_addr = lease -> ip_addr;
 	lt -> starts = lease -> starts;
 	lt -> ends = lease -> ends;
+	lt -> timestamp = lease -> timestamp;
 	lt -> uid_len = lease -> uid_len;
 	lt -> uid_max = lease -> uid_max;
 	if (lease -> uid == lease -> uid_buf) {
@@ -1416,15 +1326,14 @@ int lease_copy (struct lease **lp,
 		executable_statement_reference (&lt -> on_release,
 						lease -> on_release,
 						file, line);
-	lt->flags = lease->flags;
-	lt->tstp = lease->tstp;
-	lt->tsfp = lease->tsfp;
-	lt->atsfp = lease->atsfp;
-	lt->cltt = lease -> cltt;
-	lt->binding_state = lease->binding_state;
-	lt->next_binding_state = lease->next_binding_state;
-	status = lease_reference(lp, lt, file, line);
-	lease_dereference(&lt, MDL);
+	lt -> flags = lease -> flags;
+	lt -> tstp = lease -> tstp;
+	lt -> tsfp = lease -> tsfp;
+	lt -> cltt = lease -> cltt;
+	lt -> binding_state = lease -> binding_state;
+	lt -> next_binding_state = lease -> next_binding_state;
+	status = lease_reference (lp, lt, file, line);
+	lease_dereference (&lt, MDL);
 	return status == ISC_R_SUCCESS;
 }
 
@@ -1467,14 +1376,7 @@ void release_lease (lease, packet)
 		/* Blow away any bindings. */
 		if (lease -> scope)
 			binding_scope_dereference (&lease -> scope, MDL);
-
-		/* Set sort times to the present. */
 		lease -> ends = cur_time;
-		/* Lower layers of muckery set tstp to ->ends.  But we send
-		 * protocol messages before this.  So it is best to set
-		 * tstp now anyway.
-		 */
-		lease->tstp = cur_time;
 #if defined (FAILOVER_PROTOCOL)
 		if (lease -> pool && lease -> pool -> failover_peer) {
 			lease -> next_binding_state = FTS_RELEASED;
@@ -1496,16 +1398,10 @@ void abandon_lease (lease, message)
 	const char *message;
 {
 	struct lease *lt = (struct lease *)0;
-#if defined (NSUPDATE)
-	ddns_removals (lease);
-#endif
 
 	if (!lease_copy (&lt, lease, MDL))
 		return;
 
-	if (lt->scope)
-		binding_scope_dereference(&lt->scope, MDL);
-
 	lt -> ends = cur_time; /* XXX */
 	lt -> next_binding_state = FTS_ABANDONED;
 
@@ -1528,9 +1424,6 @@ void dissociate_lease (lease)
 	struct lease *lease;
 {
 	struct lease *lt = (struct lease *)0;
-#if defined (NSUPDATE)
-	ddns_removals (lease);
-#endif
 
 	if (!lease_copy (&lt, lease, MDL))
 		return;
@@ -1563,26 +1456,24 @@ void pool_timer (vpool)
 	struct lease *lt = (struct lease *)0;
 	struct lease *next = (struct lease *)0;
 	struct lease *lease = (struct lease *)0;
-#define FREE_LEASES 0
-#define ACTIVE_LEASES 1
-#define EXPIRED_LEASES 2
-#define ABANDONED_LEASES 3
-#define BACKUP_LEASES 4
-#define RESERVED_LEASES 5
-	struct lease **lptr[RESERVED_LEASES+1];
+	struct lease **lptr [5];
 	TIME next_expiry = MAX_TIME;
 	int i;
 
 	pool = (struct pool *)vpool;
 
+#define FREE_LEASES 0
 	lptr [FREE_LEASES] = &pool -> free;
+#define ACTIVE_LEASES 1
 	lptr [ACTIVE_LEASES] = &pool -> active;
+#define EXPIRED_LEASES 2
 	lptr [EXPIRED_LEASES] = &pool -> expired;
+#define ABANDONED_LEASES 3
 	lptr [ABANDONED_LEASES] = &pool -> abandoned;
+#define BACKUP_LEASES 4
 	lptr [BACKUP_LEASES] = &pool -> backup;
-	lptr[RESERVED_LEASES] = &pool->reserved;
 
-	for (i = FREE_LEASES; i <= RESERVED_LEASES; i++) {
+	for (i = FREE_LEASES; i <= BACKUP_LEASES; i++) {
 		/* If there's nothing on the queue, skip it. */
 		if (!*(lptr [i]))
 			continue;
@@ -1653,8 +1544,8 @@ void pool_timer (vpool)
 int find_lease_by_ip_addr (struct lease **lp, struct iaddr addr,
 			   const char *file, int line)
 {
-	return lease_ip_hash_lookup(lp, lease_ip_addr_hash, addr.iabuf,
-				    addr.len, file, line);
+	return lease_hash_lookup (lp, lease_ip_addr_hash,
+				  addr.iabuf, addr.len, file, line);
 }
 
 int find_lease_by_uid (struct lease **lp, const unsigned char *uid,
@@ -1662,7 +1553,7 @@ int find_lease_by_uid (struct lease **lp, const unsigned char *uid,
 {
 	if (len == 0)
 		return 0;
-	return lease_id_hash_lookup (lp, lease_uid_hash, uid, len, file, line);
+	return lease_hash_lookup (lp, lease_uid_hash, uid, len, file, line);
 }
 
 int find_lease_by_hw_addr (struct lease **lp,
@@ -1671,121 +1562,33 @@ int find_lease_by_hw_addr (struct lease **lp,
 {
 	if (hwlen == 0)
 		return 0;
-	return lease_id_hash_lookup(lp, lease_hw_addr_hash, hwaddr, hwlen,
-				    file, line);
-}
-
-/* If the lease is preferred over the candidate, return truth.  The
- * 'cand' and 'lease' names are retained to read more clearly against
- * the 'uid_hash_add' and 'hw_hash_add' functions (this is common logic
- * to those two functions).
- *
- * 1) ACTIVE leases are preferred.  The active lease with
- *    the longest lifetime is preferred over shortest.
- * 2) "transitional states" are next, this time with the
- *    most recent CLTT.
- * 3) free/backup/etc states are next, again with CLTT.  In truth we
- *    should never see reset leases for this.
- * 4) Abandoned leases are always dead last.
- */
-static isc_boolean_t
-client_lease_preferred(struct lease *cand, struct lease *lease)
-{
-	if (cand->binding_state == FTS_ACTIVE) {
-		if (lease->binding_state == FTS_ACTIVE &&
-		    lease->ends >= cand->ends)
-			return ISC_TRUE;
-	} else if (cand->binding_state == FTS_EXPIRED ||
-		   cand->binding_state == FTS_RELEASED) {
-		if (lease->binding_state == FTS_ACTIVE)
-			return ISC_TRUE;
-
-		if ((lease->binding_state == FTS_EXPIRED ||
-		     lease->binding_state == FTS_RELEASED) &&
-		    lease->cltt >= cand->cltt)
-			return ISC_TRUE;
-	} else if (cand->binding_state != FTS_ABANDONED) {
-		if (lease->binding_state == FTS_ACTIVE ||
-		    lease->binding_state == FTS_EXPIRED ||
-		    lease->binding_state == FTS_RELEASED)
-			return ISC_TRUE;
-
-		if (lease->binding_state != FTS_ABANDONED &&
-		    lease->cltt >= cand->cltt)
-			return ISC_TRUE;
-	} else /* (cand->binding_state == FTS_ABANDONED) */ {
-		if (lease->binding_state != FTS_ABANDONED ||
-		    lease->cltt >= cand->cltt)
-			return ISC_TRUE;
-	}
-
-	return ISC_FALSE;
+	return lease_hash_lookup (lp, lease_hw_addr_hash,
+				  hwaddr, hwlen, file, line);
 }
 
 /* Add the specified lease to the uid hash. */
-void
-uid_hash_add(struct lease *lease)
+
+void uid_hash_add (lease)
+	struct lease *lease;
 {
-	struct lease *head = NULL;
-	struct lease *cand = NULL;
-	struct lease *prev = NULL;
-	struct lease *next = NULL;
+	struct lease *head = (struct lease *)0;
+	struct lease *next = (struct lease *)0;
+
 
 	/* If it's not in the hash, just add it. */
 	if (!find_lease_by_uid (&head, lease -> uid, lease -> uid_len, MDL))
-		lease_id_hash_add(lease_uid_hash, lease->uid, lease->uid_len,
-				  lease, MDL);
+		lease_hash_add (lease_uid_hash, lease -> uid,
+				lease -> uid_len, lease, MDL);
 	else {
-		/* Otherwise, insert it into the list in order of its
-		 * preference for "resuming allocation to the client."
-		 *
-		 * Because we don't have control of the hash bucket index
-		 * directly, we have to remove and re-insert the client
-		 * id into the hash if we're inserting onto the head.
-		 */
-		lease_reference(&cand, head, MDL);
-		while (cand != NULL) {
-			if (client_lease_preferred(cand, lease))
-				break;
-
-			if (prev != NULL)
-				lease_dereference(&prev, MDL);
-			lease_reference(&prev, cand, MDL);
-
-			if (cand->n_uid != NULL)
-				lease_reference(&next, cand->n_uid, MDL);
-
-			lease_dereference(&cand, MDL);
-
-			if (next != NULL) {
-				lease_reference(&cand, next, MDL);
-				lease_dereference(&next, MDL);
-			}
+		/* Otherwise, attach it to the end of the list. */
+		while (head -> n_uid) {
+			lease_reference (&next, head -> n_uid, MDL);
+			lease_dereference (&head, MDL);
+			lease_reference (&head, next, MDL);
+			lease_dereference (&next, MDL);
 		}
-
-		/* If we want to insert 'before cand', and prev is NULL,
-		 * then it was the head of the list.  Assume that position.
-		 */
-		if (prev == NULL) {
-			lease_reference(&lease->n_uid, head, MDL);
-			lease_id_hash_delete(lease_uid_hash, lease->uid,
-					     lease->uid_len, MDL);
-			lease_id_hash_add(lease_uid_hash, lease->uid,
-					  lease->uid_len, lease, MDL);
-		} else /* (prev != NULL) */ {
-			if(prev->n_uid != NULL) {
-				lease_reference(&lease->n_uid, prev->n_uid,
-						MDL);
-				lease_dereference(&prev->n_uid, MDL);
-			}
-			lease_reference(&prev->n_uid, lease, MDL);
-
-			lease_dereference(&prev, MDL);
-		}
-
-		if (cand != NULL)
-			lease_dereference(&cand, MDL);
-		lease_dereference(&head, MDL);
+		lease_reference (&head -> n_uid, lease, MDL);
+		lease_dereference (&head, MDL);
 	}
 }
 
@@ -1808,12 +1611,13 @@ void uid_hash_delete (lease)
 	   remove the hash table entry and add a new one with the
 	   next lease on the list (if there is one). */
 	if (head == lease) {
-		lease_id_hash_delete(lease_uid_hash, lease->uid,
-				     lease->uid_len, MDL);
+		lease_hash_delete (lease_uid_hash,
+				   lease -> uid, lease -> uid_len, MDL);
 		if (lease -> n_uid) {
-			lease_id_hash_add(lease_uid_hash, lease->n_uid->uid,
-					  lease->n_uid->uid_len, lease->n_uid,
-					  MDL);
+			lease_hash_add (lease_uid_hash,
+					lease -> n_uid -> uid,
+					lease -> n_uid -> uid_len,
+					lease -> n_uid, MDL);
 			lease_dereference (&lease -> n_uid, MDL);
 		}
 	} else {
@@ -1838,74 +1642,30 @@ void uid_hash_delete (lease)
 
 /* Add the specified lease to the hardware address hash. */
 
-void
-hw_hash_add(struct lease *lease)
+void hw_hash_add (lease)
+	struct lease *lease;
 {
-	struct lease *head = NULL;
-	struct lease *cand = NULL;
-	struct lease *prev = NULL;
-	struct lease *next = NULL;
+	struct lease *head = (struct lease *)0;
+	struct lease *next = (struct lease *)0;
 
 	/* If it's not in the hash, just add it. */
 	if (!find_lease_by_hw_addr (&head, lease -> hardware_addr.hbuf,
 				    lease -> hardware_addr.hlen, MDL))
-		lease_id_hash_add(lease_hw_addr_hash,
-				  lease->hardware_addr.hbuf,
-				  lease->hardware_addr.hlen, lease, MDL);
+		lease_hash_add (lease_hw_addr_hash,
+				lease -> hardware_addr.hbuf,
+				lease -> hardware_addr.hlen,
+				lease, MDL);
 	else {
-		/* Otherwise, insert it into the list in order of its
-		 * preference for "resuming allocation to the client."
-		 *
-		 * Because we don't have control of the hash bucket index
-		 * directly, we have to remove and re-insert the client
-		 * id into the hash if we're inserting onto the head.
-		 */
-		lease_reference(&cand, head, MDL);
-		while (cand != NULL) {
-			if (client_lease_preferred(cand, lease))
-				break;
-
-			if (prev != NULL)
-				lease_dereference(&prev, MDL);
-			lease_reference(&prev, cand, MDL);
-
-			if (cand->n_hw != NULL)
-				lease_reference(&next, cand->n_hw, MDL);
-
-			lease_dereference(&cand, MDL);
-
-			if (next != NULL) {
-				lease_reference(&cand, next, MDL);
-				lease_dereference(&next, MDL);
-			}
+		/* Otherwise, attach it to the end of the list. */
+		while (head -> n_hw) {
+			lease_reference (&next, head -> n_hw, MDL);
+			lease_dereference (&head, MDL);
+			lease_reference (&head, next, MDL);
+			lease_dereference (&next, MDL);
 		}
 
-		/* If we want to insert 'before cand', and prev is NULL,
-		 * then it was the head of the list.  Assume that position.
-		 */
-		if (prev == NULL) {
-			lease_reference(&lease->n_hw, head, MDL);
-			lease_id_hash_delete(lease_hw_addr_hash,
-					     lease->hardware_addr.hbuf,
-					     lease->hardware_addr.hlen, MDL);
-			lease_id_hash_add(lease_hw_addr_hash,
-					  lease->hardware_addr.hbuf,
-					  lease->hardware_addr.hlen,
-					  lease, MDL);
-		} else /* (prev != NULL) */ {
-			if(prev->n_hw != NULL) {
-				lease_reference(&lease->n_hw, prev->n_hw,
-						MDL);
-				lease_dereference(&prev->n_hw, MDL);
-			}
-			lease_reference(&prev->n_hw, lease, MDL);
-
-			lease_dereference(&prev, MDL);
-		}
-
-		if (cand != NULL)
-			lease_dereference(&cand, MDL);
-		lease_dereference(&head, MDL);
+		lease_reference (&head -> n_hw, lease, MDL);
+		lease_dereference (&head, MDL);
 	}
 }
 
@@ -1929,15 +1689,15 @@ void hw_hash_delete (lease)
 	   remove the hash table entry and add a new one with the
 	   next lease on the list (if there is one). */
 	if (head == lease) {
-		lease_id_hash_delete(lease_hw_addr_hash,
-				     lease->hardware_addr.hbuf,
-				     lease->hardware_addr.hlen, MDL);
-		if (lease->n_hw) {
-			lease_id_hash_add(lease_hw_addr_hash,
-					  lease->n_hw->hardware_addr.hbuf,
-					  lease->n_hw->hardware_addr.hlen,
-					  lease->n_hw, MDL);
-			lease_dereference(&lease->n_hw, MDL);
+		lease_hash_delete (lease_hw_addr_hash,
+				   lease -> hardware_addr.hbuf,
+				   lease -> hardware_addr.hlen, MDL);
+		if (lease -> n_hw) {
+			lease_hash_add (lease_hw_addr_hash,
+					lease -> n_hw -> hardware_addr.hbuf,
+					lease -> n_hw -> hardware_addr.hlen,
+					lease -> n_hw, MDL);
+			lease_dereference (&lease -> n_hw, MDL);
 		}
 	} else {
 		/* Otherwise, look for the lease in the list of leases
@@ -1974,29 +1734,10 @@ int write_leases ()
 	struct host_decl *hp;
 	struct group_object *gp;
 	struct hash_bucket *hb;
-	struct class *cp;
-	struct collection *colp;
 	int i;
 	int num_written;
-	struct lease **lptr[RESERVED_LEASES+1];
+	struct lease **lptr [5];
 
-	/* write all the dynamically-created class declarations. */
-	if (collections->classes) {
-		numclasseswritten = 0;
-		for (colp = collections ; colp ; colp = colp->next) {
-			for (cp = colp->classes ; cp ; cp = cp->nic) {
-				write_named_billing_class(
-						(unsigned char *)cp->name,
-							  0, cp);
-			}
-		}
-
-		/* XXXJAB this number doesn't include subclasses... */ 
-		log_info ("Wrote %d class decls to leases file.",
-			  numclasseswritten);
-	}
-	
-			
 	/* Write all the dynamically-created group declarations. */
 	if (group_name_hash) {
 	    num_written = 0;
@@ -2067,9 +1808,8 @@ int write_leases ()
 		lptr [EXPIRED_LEASES] = &p -> expired;
 		lptr [ABANDONED_LEASES] = &p -> abandoned;
 		lptr [BACKUP_LEASES] = &p -> backup;
-		lptr [RESERVED_LEASES] = &p->reserved;
 
-		for (i = FREE_LEASES; i <= RESERVED_LEASES; i++) {
+		for (i = FREE_LEASES; i <= BACKUP_LEASES; i++) {
 		    for (l = *(lptr [i]); l; l = l -> next) {
 #if !defined (DEBUG_DUMP_ALL_LEASES)
 			if (l -> hardware_addr.hlen ||
@@ -2091,23 +1831,9 @@ int write_leases ()
 	return 1;
 }
 
-/* In addition to placing this lease upon a lease queue depending on its
- * state, it also keeps track of the number of FREE and BACKUP leases in
- * existence, and sets the sort_time on the lease.
- *
- * Sort_time is used in pool_timer() to determine when the lease will
- * bubble to the top of the list and be supersede_lease()'d into its next
- * state (possibly, if all goes well).  Example, ACTIVE leases move to
- * EXPIRED state when the 'ends' value is reached, so that is its sort
- * time.  Most queues are sorted by 'ends', since it is generally best
- * practice to re-use the oldest lease, to reduce address collision
- * chances.
- */
 int lease_enqueue (struct lease *comp)
 {
 	struct lease **lq, *prev, *lp;
-	static struct lease **last_lq = NULL;
-	static struct lease *last_insert_point = NULL;
 
 	/* No queue to put it on? */
 	if (!comp -> pool)
@@ -2116,12 +1842,8 @@ int lease_enqueue (struct lease *comp)
 	/* Figure out which queue it's going to. */
 	switch (comp -> binding_state) {
 	      case FTS_FREE:
-		if (comp->flags & RESERVED_LEASE) {
-			lq = &comp->pool->reserved;
-		} else {
-			lq = &comp->pool->free;
-			comp->pool->free_leases++;
-		}
+		lq = &comp -> pool -> free;
+		comp -> pool -> free_leases++;
 		comp -> sort_time = comp -> ends;
 		break;
 
@@ -2134,26 +1856,7 @@ int lease_enqueue (struct lease *comp)
 	      case FTS_RELEASED:
 	      case FTS_RESET:
 		lq = &comp -> pool -> expired;
-#if defined(FAILOVER_PROTOCOL)
-		/* In partner_down, tsfp is the time at which the lease
-		 * may be reallocated (stos+mclt).  We can do that with
-		 * lease_mine_to_reallocate() anywhere between tsfp and
-		 * ends.  But we prefer to wait until ends before doing it
-		 * automatically (choose the greater of the two).  Note
-		 * that 'ends' is usually a historic timestamp in the
-		 * case of expired leases, is really only in the future
-		 * on released leases, and if we know a lease to be released
-		 * the peer might still know it to be active...in which case
-		 * it's possible the peer has renewed this lease, so avoid
-		 * doing that.
-		 */
-		if (comp->pool->failover_peer &&
-		    comp->pool->failover_peer->me.state == partner_down)
-			comp->sort_time = (comp->tsfp > comp->ends) ?
-					  comp->tsfp : comp->ends;
-		else
-#endif
-			comp->sort_time = comp->ends;
+		comp -> sort_time = comp -> ends;
 
 		break;
 
@@ -2163,12 +1866,8 @@ int lease_enqueue (struct lease *comp)
 		break;
 
 	      case FTS_BACKUP:
-		if (comp->flags & RESERVED_LEASE) {
-			lq = &comp->pool->reserved;
-		} else {
-			lq = &comp->pool->backup;
-			comp->pool->backup_leases++;
-		}
+		lq = &comp -> pool -> backup;
+		comp -> pool -> backup_leases++;
 		comp -> sort_time = comp -> ends;
 		break;
 
@@ -2181,28 +1880,13 @@ int lease_enqueue (struct lease *comp)
 		return 0;
 	}
 
-	/* This only works during server startup: during runtime, the last
-	 * lease may be dequeued inbetween calls.  If the queue is the same
-	 * as was used previously, and the lease structure isn't (this is not
-	 * a re-queue), use that as a starting point for the insertion-sort.
-	 */
-	if ((server_starting & SS_QFOLLOW) && (lq == last_lq) &&
-	    (comp != last_insert_point) && 
-	    (last_insert_point->sort_time <= comp->sort_time)) {
-		prev = last_insert_point;
-		lp = prev->next;
-	} else {
-		prev = NULL;
-		lp = *lq;
-	}
-
 	/* Insertion sort the lease onto the appropriate queue. */
-	for (; lp ; lp = lp->next) {
+	prev = (struct lease *)0;
+	for (lp = *lq; lp; lp = lp -> next) {
 		if (lp -> sort_time >= comp -> sort_time)
 			break;
 		prev = lp;
 	}
-
 	if (prev) {
 		if (prev -> next) {
 			lease_reference (&comp -> next, prev -> next, MDL);
@@ -2216,8 +1900,6 @@ int lease_enqueue (struct lease *comp)
 		}
 		lease_reference (lq, comp, MDL);
 	}
-	last_insert_point = comp;
-	last_lq = lq;
 	return 1;
 }
 
@@ -2225,39 +1907,33 @@ int lease_enqueue (struct lease *comp)
    in each appropriate hash, understanding that it's already by definition
    in lease_ip_addr_hash. */
 
-isc_result_t
-lease_instantiate(const void *key, unsigned len, void *object)
+void lease_instantiate (const unsigned char *val, unsigned len,
+			struct lease *lease)
 {
-	struct lease *lease = object;
 	struct class *class;
 	/* XXX If the lease doesn't have a pool at this point, it's an
 	   XXX orphan, which we *should* keep around until it expires,
 	   XXX but which right now we just forget. */
 	if (!lease -> pool) {
-		lease_ip_hash_delete(lease_ip_addr_hash, lease->ip_addr.iabuf,
-				     lease->ip_addr.len, MDL);
-		return ISC_R_SUCCESS;
+		lease_hash_delete (lease_ip_addr_hash,
+				   lease -> ip_addr.iabuf,
+				   lease -> ip_addr.len, MDL);
+		return;
 	}
 		
-	/* Put the lease on the right queue.  Failure to queue is probably
-	 * due to a bogus binding state.  In such a case, we claim success,
-	 * so that later leases in a hash_foreach are processed, but we
-	 * return early as we really don't want hw address hash entries or
-	 * other cruft to surround such a bogus entry.
-	 */
-	if (!lease_enqueue(lease))
-		return ISC_R_SUCCESS;
+	/* Put the lease on the right queue. */
+	lease_enqueue (lease);
 
 	/* Record the lease in the uid hash if possible. */
 	if (lease -> uid) {
 		uid_hash_add (lease);
 	}
-
+	
 	/* Record it in the hardware address hash if possible. */
 	if (lease -> hardware_addr.hlen) {
 		hw_hash_add (lease);
 	}
-
+	
 	/* If the lease has a billing class, set up the billing. */
 	if (lease -> billing_class) {
 		class = (struct class *)0;
@@ -2272,7 +1948,7 @@ lease_instantiate(const void *key, unsigned len, void *object)
 			bill_class (lease, class);
 		class_dereference (&class, MDL);
 	}
-	return ISC_R_SUCCESS;
+	return;
 }
 
 /* Run expiry events on every pool.   This is called on startup so that
@@ -2287,22 +1963,14 @@ void expire_all_pools ()
 	struct hash_bucket *hb;
 	int i;
 	struct lease *l;
-	struct lease **lptr[RESERVED_LEASES+1];
-
-	/* Indicate that we are in the startup phase */
-	server_starting = SS_NOSYNC | SS_QFOLLOW;
+	struct lease **lptr [5];
 
 	/* First, go over the hash list and actually put all the leases
 	   on the appropriate lists. */
-	lease_ip_hash_foreach(lease_ip_addr_hash, lease_instantiate);
+	lease_hash_foreach (lease_ip_addr_hash, lease_instantiate);
 
 	/* Loop through each pool in each shared network and call the
-	 * expiry routine on the pool.  It is no longer safe to follow
-	 * the queue insertion point, as expiration of a lease can move
-	 * it between queues (and this may be the lease that function
-	 * points at).
-	 */
-	server_starting &= ~SS_QFOLLOW;
+	   expiry routine on the pool. */
 	for (s = shared_networks; s; s = s -> next) {
 	    for (p = s -> pools; p; p = p -> next) {
 		pool_timer (p);
@@ -2310,35 +1978,25 @@ void expire_all_pools ()
 		p -> lease_count = 0;
 		p -> free_leases = 0;
 		p -> backup_leases = 0;
-
+		
 		lptr [FREE_LEASES] = &p -> free;
 		lptr [ACTIVE_LEASES] = &p -> active;
 		lptr [EXPIRED_LEASES] = &p -> expired;
 		lptr [ABANDONED_LEASES] = &p -> abandoned;
 		lptr [BACKUP_LEASES] = &p -> backup;
-		lptr [RESERVED_LEASES] = &p->reserved;
 
-		for (i = FREE_LEASES; i <= RESERVED_LEASES; i++) {
+		for (i = FREE_LEASES; i <= BACKUP_LEASES; i++) {
 		    for (l = *(lptr [i]); l; l = l -> next) {
 			p -> lease_count++;
 			if (l -> ends <= cur_time) {
-				if (l->binding_state == FTS_FREE) {
-					if (i == FREE_LEASES)
-						p->free_leases++;
-					else if (i != RESERVED_LEASES)
-						log_fatal("Impossible case "
-							  "at %s:%d.", MDL);
-				} else if (l->binding_state == FTS_BACKUP) {
-					if (i == BACKUP_LEASES)
-						p->backup_leases++;
-					else if (i != RESERVED_LEASES)
-						log_fatal("Impossible case "
-							  "at %s:%d.", MDL);
-				}
+				if (l -> binding_state == FTS_FREE)
+					p -> free_leases++;
+				else if (l -> binding_state == FTS_BACKUP)
+					p -> backup_leases++;
 			}
 #if defined (FAILOVER_PROTOCOL)
 			if (p -> failover_peer &&
-			    l -> tstp > l -> atsfp &&
+			    l -> tstp > l -> tsfp &&
 			    !(l -> flags & ON_UPDATE_QUEUE)) {
 				l -> desired_binding_state = l -> binding_state;
 				dhcp_failover_queue_update (l, 1);
@@ -2348,9 +2006,6 @@ void expire_all_pools ()
 		}
 	    }
 	}
-
-	/* turn off startup phase */
-	server_starting = 0;
 }
 
 void dump_subnets ()
@@ -2359,7 +2014,7 @@ void dump_subnets ()
 	struct shared_network *s;
 	struct subnet *n;
 	struct pool *p;
-	struct lease **lptr[RESERVED_LEASES+1];
+	struct lease **lptr [5];
 	int i;
 
 	log_info ("Subnets:");
@@ -2377,9 +2032,8 @@ void dump_subnets ()
 		lptr [EXPIRED_LEASES] = &p -> expired;
 		lptr [ABANDONED_LEASES] = &p -> abandoned;
 		lptr [BACKUP_LEASES] = &p -> backup;
-		lptr [RESERVED_LEASES] = &p->reserved;
 
-		for (i = FREE_LEASES; i <= RESERVED_LEASES; i++) {
+		for (i = FREE_LEASES; i <= BACKUP_LEASES; i++) {
 		    for (l = *(lptr [i]); l; l = l -> next) {
 			    print_lease (l);
 		    }
@@ -2388,14 +2042,12 @@ void dump_subnets ()
 	}
 }
 
-HASH_FUNCTIONS(lease_ip, const unsigned char *, struct lease, lease_ip_hash_t,
-	       lease_reference, lease_dereference, do_ip4_hash)
-HASH_FUNCTIONS(lease_id, const unsigned char *, struct lease, lease_id_hash_t,
-	       lease_reference, lease_dereference, do_id_hash)
+HASH_FUNCTIONS (lease, const unsigned char *, struct lease, lease_hash_t,
+		lease_reference, lease_dereference)
 HASH_FUNCTIONS (host, const unsigned char *, struct host_decl, host_hash_t,
-		host_reference, host_dereference, do_string_hash)
+		host_reference, host_dereference)
 HASH_FUNCTIONS (class, const char *, struct class, class_hash_t,
-		class_reference, class_dereference, do_string_hash)
+		class_reference, class_dereference)
 
 #if defined (DEBUG_MEMORY_LEAKAGE) && \
 		defined (DEBUG_MEMORY_LEAKAGE_ON_EXIT)
@@ -2559,8 +2211,8 @@ void free_everything ()
 		if (nc -> pools) {
 		    pool_reference (&pn, nc -> pools, MDL);
 		    do {
-			struct lease **lptr[RESERVED_LEASES+1];
-
+			struct lease **lptr [5];
+			
 			if (pn) {
 			    pool_reference (&pc, pn, MDL);
 			    pool_dereference (&pn, MDL);
@@ -2569,16 +2221,15 @@ void free_everything ()
 			    pool_reference (&pn, pc -> next, MDL);
 			    pool_dereference (&pc -> next, MDL);
 			}
-
+			
 			lptr [FREE_LEASES] = &pc -> free;
 			lptr [ACTIVE_LEASES] = &pc -> active;
 			lptr [EXPIRED_LEASES] = &pc -> expired;
 			lptr [ABANDONED_LEASES] = &pc -> abandoned;
 			lptr [BACKUP_LEASES] = &pc -> backup;
-			lptr [RESERVED_LEASES] = &pc->reserved;
 
 			/* As (sigh) are leases. */
-			for (i = FREE_LEASES ; i <= RESERVED_LEASES ; i++) {
+			for (i = 0; i < 5; i++) {
 			    if (*lptr [i]) {
 				lease_reference (&ln, *lptr [i], MDL);
 				do {
